import gzip
import json

from .. import interface


class GzipDeserializer(interface.Deserializer):
    def deserialize(self, value):
        return gzip.decompress(value)


class JSONDeserializer(interface.Deserializer):
    def deserialize(self, value):
        if isinstance(value, bytes):
            value = value.decode('utf-8')
        return json.loads(value)


class NullDeserializer(interface.Deserializer):
    def deserialize(self, value):
        return value


__all__ = ['GzipDeserializer', 'NullDeserializer', 'JSONDeserializer']



import json
import os
import sqlite3
import subprocess
import unittest


class TestSQL(unittest.TestCase):
    def setUp(self):
        db_filepath = '/tmp/test1'
        with open(db_filepath, 'w'):
            pass
        self.db_filepath = db_filepath
        db = sqlite3.connect(db_filepath)
        c = db.cursor()
        c.execute('CREATE TABLE test_table1 (col1 text, col2 int);')
        c.execute("INSERT INTO test_table1 (col1, col2) VALUES ('test', 1);")
        c.execute("INSERT INTO test_table1 (col1, col2) VALUES ('test', 2);")
        c.execute("INSERT INTO test_table1 (col1, col2) VALUES ('test', 3);")
        db.commit()
        db.close()

    def tearDown(self):
        os.unlink(self.db_filepath)

    def run_recipe(self, taskname):
        c = subprocess.run([
            'python', '-m',
            'bamboo_crawler',
            '-r', 'tests/sql/recipe.yml',
            '-t', taskname],
            stdout=subprocess.PIPE)
        self.assertEqual(c.returncode, 0)
        return c

    def test_test1(self):
        c = self.run_recipe('test1')
        j = json.loads(c.stdout)
        self.assertEqual(j['col1'], 'test')
        self.assertEqual(j['col2'], 1)

    def test_test2(self):
        db = sqlite3.connect(self.db_filepath)
        c1, = db.execute('SELECT COUNT(*) FROM test_table1').fetchone()
        self.run_recipe('test2')
        db = sqlite3.connect(self.db_filepath)
        c2, = db.execute('SELECT COUNT(*) FROM test_table1').fetchone()
        self.assertEqual(c1 + 1, c2)

    def test_test3(self):
        c = self.run_recipe('test3')
        j = json.loads(c.stdout)
        self.assertEqual(j['col1'], 'test')
        self.assertEqual(j['col2'], 1)

    def test_test4(self):
        db = sqlite3.connect(self.db_filepath)
        c1, = db.execute('SELECT COUNT(*) FROM test_table1').fetchone()
        self.run_recipe('test4')
        db = sqlite3.connect(self.db_filepath)
        c2, = db.execute('SELECT COUNT(*) FROM test_table1').fetchone()
        self.assertEqual(c1 + 1, c2)



import hashlib
import json
import sys
from types import MappingProxyType

import boto3
import sqlalchemy
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.sql import text

from .. import interface


class StdoutOutputter(interface.Outputter):
    def write(self, value, *, context=None):
        if isinstance(value, bytes):
            try:
                enc = sys.getdefaultencoding()
                value = value.decode(enc)
            except Exception:
                pass
        sys.stdout.write(value)


class SQSS3Outputter(interface.Outputter):

    def __init__(self, bucket_name, queue_name,
                 *,
                 s3_config=MappingProxyType({}),
                 sqs_config=MappingProxyType({})):
        super().__init__()
        s3 = boto3.resource('s3', **s3_config)
        sqs = boto3.resource('sqs', **sqs_config)
        bucket = s3.Bucket(bucket_name)
        queue = sqs.get_queue_by_name(QueueName=queue_name)
        self.bucket = bucket
        self.queue = queue

    def write(self, value, *, context=None):
        bucket = self.bucket
        queue = self.queue
        encoding = sys.getdefaultencoding()
        if context is not None:
            metadata = context.metadata
        else:
            metadata = {}
        if isinstance(value, str):
            value = value.encode(encoding)

        name = hashlib.sha256(value).hexdigest()
        obj = bucket.Object(name)
        obj.put(
            Body=value,
            ContentEncoding=encoding,
            ContentType='text/plain',
        )
        j = json.dumps({'s3_body': obj.key, 'metadata': metadata})
        queue.send_message(MessageBody=j)


class SQLOutputter(interface.Outputter):
    def __init__(self, url, *, table=None, query=None):
        if query is None and table is None:
            raise NotImplementedError
        if query is not None and table is not None:
            raise NotImplementedError
        self.engine = sqlalchemy.create_engine(url)
        self._type = 'query'
        if table is not None:
            self._type = 'table'
            metadata = sqlalchemy.MetaData()
            metadata.reflect(self.engine, only=[table])
            Base = automap_base(metadata=metadata)
            Base.prepare()
            query = metadata.tables[table].insert()
        self.query = query

    def write(self, value, *, context=None):
        j = json.loads(value)
        if self._type == 'table':
            self.engine.execute(self.query.values(j))
        else:
            self.engine.execute(text(self.query), j)


__all__ = ['StdoutOutputter', 'SQSS3Outputter', 'SQLOutputter']



import time


class Task(object):

    def __init__(self, *,
                 name,
                 inputter,
                 processor,
                 outputter,
                 deserializer=None,
                 serializer=None):
        self.name = name
        self.inputter = inputter
        self.processor = processor
        self.outputter = outputter
        self.deserializer = deserializer
        self.serializer = serializer

    def do(self):
        context = self.inputter.read()
        body = context.body
        if self.deserializer is not None:
            body = self.deserializer.deserialize(body)
        for d in self.processor.process(body, context=context):
            job_name = self.name
            class_name = self.processor.__class__.__name__
            metadatakey = 'processed_time_{}_{}'.format(job_name, class_name)
            timestamp = int(time.time())
            context.add_metadata(**{metadatakey: timestamp})
            if self.serializer is not None:
                d = self.serializer.serialize(d)
            self.outputter.write(d, context=context)
        self.inputter.done()



from abc import ABCMeta
from abc import abstractmethod
from typing import Any


class Inputter(metaclass=ABCMeta):
    @abstractmethod
    def read(self):
        ...

    def done(self):
        pass


class Outputter(metaclass=ABCMeta):
    @abstractmethod
    def write(self, value, *, context=None):
        ...


class Processor(metaclass=ABCMeta):
    @abstractmethod
    def process(self, value, *, context=None):
        yield ...


class Crawler(Processor):
    def process(self, value, *, context=None):
        for content in self.crawl(value, context=context):
            yield content

    @abstractmethod
    def crawl(self, location, *, context=None):
        yield ...


class Scraper(Processor):
    def process(self, value, *, context=None):
        for content in self.scrape(value, context=None):
            yield content

    @abstractmethod
    def scrape(self, data, *, context=None):
        yield ...


class Deserializer(metaclass=ABCMeta):

    @abstractmethod
    def deserialize(self, data: bytes):
        ...


class Serializer(metaclass=ABCMeta):

    @abstractmethod
    def serialize(self, value: Any):
        ...


class Context(object):
    def __init__(self, body, **kwargs):
        self.body = body
        self.metadata = kwargs.copy()

    def add_metadata(self, **kwargs):
        self.metadata.update(kwargs)



import subprocess
import unittest


class TestCrawling(unittest.TestCase):
    def setUp(self):
        server = subprocess.Popen([
            'python',
            '-m', 'http.server',
            '8000'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd='tests/crawling/assets')
        self.server = server

    def tearDown(self):
        server = self.server
        server.stderr.close()
        server.stdout.close()
        server.kill()
        server.wait()

    def run_recipe(self, taskname):
        c = subprocess.run([
            'python', '-m',
            'bamboo_crawler',
            '-r', 'tests/crawling/recipe.yml',
            '-t', taskname],
            stdout=subprocess.PIPE)
        self.assertEqual(c.returncode, 0)
        return c

    def test_something(self):
        c = self.run_recipe('index_html')
        self.assertEqual(c.stdout, b'<body>test!test!test!</body>\n')



import json

from .. import interface


class NullSerializer(interface.Serializer):
    def serialize(self, value):
        return value


class JSONSerializer(interface.Serializer):
    def serialize(self, value):
        return json.dumps(value)


__all__ = ['NullSerializer', 'JSONSerializer']
